// Audio Init
#include <notes.h>
#define FASTLED_INTERNAL
#include <FastLED.h>
#define NUM_LEDS 169
const int musicPin = 9;

int pacman[] = {
  NOTE_B4, 1, NOTE_B5, 1, NOTE_FS5, 1, NOTE_DS5, 1, NOTE_B5, 1, NOTE_FS5, 1, NOTE_DS5, 0.5, NOTE_C5, 1, NOTE_C6, 1, NOTE_G5, 1, NOTE_E5, 1, NOTE_C6, 1, NOTE_G5, 1, NOTE_E5, 0.5, NOTE_B4, 1, NOTE_B5, 1, NOTE_FS5, 1, NOTE_DS5, 1, NOTE_B5, 1, NOTE_FS5, 1, NOTE_DS5, 0.5, NOTE_DS5, 2, NOTE_E5, 1, NOTE_F5, 1, NOTE_F5, 2, NOTE_FS5, 1, NOTE_G5, 1, NOTE_G5, 2, NOTE_GS5, 1, NOTE_A5, 1, NOTE_B5, 0.5
};
int pacmanlength = sizeof(pacman);

int pogsong[] = {
  NOTE_B4, 1, 0, 1, NOTE_B4, 1, 0, 1, NOTE_B4, 1, NOTE_A4, 1, 0, 1, NOTE_B4, 1, 0, 1, NOTE_D5, 1, 0, 1, NOTE_B4, 1, 0, 1, NOTE_A4, 1, NOTE_AS4, 0, 1, NOTE_B4, 1, 0, 1, NOTE_B4, 1, 0, 1, NOTE_B4, 1, NOTE_A4, 1
};

int poglength = sizeof(pogsong);

void playSong(int song[], int musicPin, int lengthA, float bpm) {
  float spb = (1/(bpm / 60)) * 1000;
  for (int i = 0; i < lengthA; i+=2) {
    float noteDuration = spb * song[i+1];
    tone(musicPin, song[i], noteDuration);
    delay(noteDuration + 50);
    noTone(musicPin);
  }
  //digitalWrite(musicPin, HIGH);
}

// Screen Init

CRGB leds[NUM_LEDS];
const int gridSize = 13;
int currScreen[gridSize][gridSize];
bool inGame = false; //make false for menu to appear

CRGB colors[] = {CRGB(0,0,0), CRGB(0,0,110), CRGB(30,30,30), CRGB(100,200,100), CRGB(150,30,30), CRGB(255,255,0), CRGB(255,0,0), CRGB(250,0,150), CRGB(0,255,255), CRGB(255,150,0)};
/* 0 = off
* 1 = wall
* 2 = pellet
* 3 = power pellet
* 4 = cherries
* 5 = pacman
* rest = ghosts
*/

void setScreen(int level[13][13])
{

  for(int i = 0; i<13;i++){
    for (int j = 0; j<13; j++){
      int m = i*13+j;
      int n = floor(m/13);
      int o = m;
      if(n%2==1){
        o = 13*n + (13-(m+1-13*n));
      }
      leds[o] = colors[level[i][j]];
    }
  }
  
  FastLED.show();
}

//Game init
int pacPos[2] = {1,1};
int ghosts[3][6] = {{1,3,0,1,8,-1},
                   {9,3,1,0,9,-1},
                   {11,11,0,-1,7,-1}}; 

int startGhosts[3][6] = {{1,3,0,1,8,-1},
                   {3,3,1,0,9,-1},
                   {11,11,0,-1,7,-1}}; 
    
int powerUpTime = 0;
int numOfGhosts = 3;

int pacMap[13][13] = {{1,1,1,1,1,1,1,1,1,1,1,1,1},
                      {1,2,2,2,3,2,2,2,2,2,2,2,1},
                      {1,2,1,1,1,2,1,2,1,1,1,2,1},
                      {1,2,1,2,2,2,1,2,2,2,1,2,1},
                      {1,2,1,2,1,2,1,2,1,2,1,2,1},
                      {1,2,2,2,1,2,1,2,1,2,2,2,1},
                      {1,2,1,1,1,2,1,2,1,1,1,2,1},
                      {1,2,1,2,2,2,1,2,2,2,1,2,1},
                      {1,2,1,2,1,1,1,1,1,2,1,2,1},
                      {1,2,2,2,2,2,2,2,2,2,2,2,1},
                      {1,2,1,1,1,1,1,1,1,1,1,2,1},
                      {1,2,2,2,2,2,2,2,2,2,2,2,1},
                      {1,1,1,1,1,1,1,1,1,1,1,1,1}};

int startPacMap[13][13] = { {1,1,1,1,1,1,1,1,1,1,1,1,1},
                            {1,3,2,2,2,2,1,2,2,2,2,3,1},
                            {1,2,1,1,1,2,1,2,1,1,1,2,1},
                            {1,2,1,2,2,2,2,2,2,2,1,2,1},
                            {1,2,1,2,1,1,2,1,1,2,1,2,1},
                            {1,2,2,2,2,1,2,1,2,2,2,2,1},
                            {1,1,1,1,2,2,2,2,2,1,1,1,1},
                            {1,2,2,2,2,1,2,1,2,2,2,2,1},
                            {1,2,1,2,1,1,2,1,1,2,1,2,1},
                            {1,2,1,2,2,2,2,2,2,2,1,2,1},
                            {1,2,1,1,1,2,1,2,1,1,1,2,1},
                            {1,3,2,2,2,2,1,2,2,2,2,3,1},
                            {1,1,1,1,1,1,1,1,1,1,1,1,1}};



void copy(int* src, int* dst, int len) {
    memcpy(dst, src, sizeof(src[0])*len);
}

void reInitGameVar(){
  for (int p = 0; p < 13; p++)
  {
    copy(startPacMap[p], pacMap[p], 13);
  }

  for (int p = 0; p < numOfGhosts; p++)
  {
    copy(startGhosts[p], ghosts[p], 6);
  }

  pacPos[0] = 6;
  pacPos[1] = 6;
  
  powerUpTime = 0;
}

bool checkCollide(int p1c, int p2c) {
  if (currScreen[p1c][p2c] != 1) {
    return true;
  }
  return false;
}

void die() {
  int dieScrn[13][13] = { {0,0,0,0,0,0,0,0,0,0,0,0,0},
                            {0,6,6,0,0,0,6,0,0,6,6,0,0},
                            {0,6,0,6,0,6,0,6,0,6,0,6,0},
                            {0,6,6,0,0,6,0,6,0,6,0,6,0},
                            {0,6,0,6,0,6,6,6,0,6,0,6,0},
                            {0,6,6,0,0,6,0,6,0,6,6,0,0},
                            {0,0,0,0,0,0,0,0,0,0,0,0,0},
                            {0,0,0,0,0,0,0,0,0,0,0,0,0},
                            {0,0,0,0,0,6,0,6,0,0,0,0,0},
                            {0,0,0,0,0,0,0,0,0,0,0,0,0},
                            {0,0,0,0,0,6,6,6,0,0,0,0,0},
                            {0,0,0,0,6,0,0,0,6,0,0,0,0},
                            {0,0,0,0,0,0,0,0,0,0,0,0,0}};
  reInitGameVar();
  setScreen(dieScrn);
  delay(3000);
  
  for (int i = 0; i < gridSize; i++)
  {
    for (int j = 0; j < gridSize; j++)
    {
      currScreen[i][j] = 4; 
    }  
  }

  

  for (int i = 0; i < gridSize; i++)
  {
    for (int j = 0; j < gridSize; j++)
    {
      currScreen[i][j] = 0; 
    }  
    delay(100);
    setScreen(currScreen);
    
  }
    
  inGame = false;
}

void win() {
  reInitGameVar();
  int winScrn[13][13] =  {{0,0,0,0,0,0,0,0,0,0,0,0,0},
                          {0,1,0,0,0,1,0,1,0,0,0,0,0},
                          {0,1,0,0,0,1,0,0,0,1,1,0,0},
                          {0,1,0,1,0,1,0,1,0,1,0,1,0},
                          {0,0,1,0,1,0,0,1,0,1,0,1,0},
                          {0,0,0,0,0,0,0,0,0,0,0,0,0},
                          {0,0,0,0,1,0,0,0,1,0,0,0,0},
                          {0,0,0,0,1,0,0,0,1,0,0,0,0},
                          {0,0,0,0,0,0,1,0,0,0,0,0,0},
                          {0,0,0,1,0,0,0,0,0,1,0,0,0},
                          {0,0,0,0,1,1,1,1,1,0,0,0,0},
                          {0,0,0,0,0,0,0,0,0,0,0,0,0}};

  

  for (int i = 0; i < gridSize; i++)
  {
    for (int j = 0; j < gridSize; j++)
    {
      currScreen[i][j] = 3; 
    }  
    delay(100);
    setScreen(currScreen);
    
  }

  reInitGameVar();
  for (int i = 0; i < gridSize; i++)
  {
    for (int j = 0; j < gridSize; j++)
    {
      currScreen[i][j] = winScrn[i][j]; 
    }  
  }
  setScreen(currScreen);
  delay(5000);

  for (int i = 0; i < gridSize; i++)
  {
    for (int j = 0; j < gridSize; j++)
    {
      currScreen[i][j] = 0; 
    }  
  }
    
  inGame = false;
}

bool levelDone() {
  for (int i = 0; i < gridSize; i++)
  {
    for (int j = 0; j < gridSize; j++)
    {
      if (pacMap[i][j] == 2){
        return false;
      }
    }  
  }
  return true;
}

void ghostAi(int & p1, int & p2, int & yVel, int & xVel) {
  int validDir[4] = {0,0,0,0};
  int key[2][4] = {{-1,1,0,0},
                   {0,0,-1,1}};
    if (checkCollide(p1 - 1, p2) && yVel != 1) {
      validDir[0] = 1;
    }
    if (checkCollide(p1 + 1, p2) && yVel != -1) {
      validDir[1] = 1;
    }
    if (checkCollide(p1, p2 - 1) && xVel != 1) {
      validDir[2] = 1;
    } 
    if (checkCollide(p1, p2 + 1) && xVel != -1) {
      validDir[3] = 1;
    }
    if (validDir[0] == 1 || validDir[1] == 1 || validDir[2] == 1 || validDir[3] == 1) {

      bool validFound = false;
      while (!validFound) {
        int randd = random(4);
        if (validDir[randd] == 1){
          validFound = true;
          yVel = key[0][randd];
          xVel = key[1][randd];
        }
      }
    }
  if (checkCollide(p1 + yVel, p2 + xVel)) {
    p1 += yVel;
    p2 += xVel;
  }
}

//Controller Init
const int xA = A0;
const int yA = A1;
const int btn = A2;
int getAxis(int axis){
  int val = map(analogRead(axis), 0, 1023, 0, 6);
  if (val > 4){
    return 1;
  }
  else if (val < 1){
    return -1;
  }
  return 0;
}
bool getBtn(){
  if(analogRead(btn) >= 500){
    return true;
  }
  else{
    return false;
  }
}

void setup() {
  // Controller setup
  pinMode(xA, INPUT);
  pinMode(yA, INPUT);
  pinMode(btn, INPUT);

  //Music setup
  pinMode(musicPin, OUTPUT);

  //Screen setup
  FastLED.addLeds<NEOPIXEL, 6> (leds, NUM_LEDS);
  FastLED.setBrightness(15);
  
  for (int p = 0; p < gridSize; p++) {
    for (int q = 0; q < gridSize; q++) {
          currScreen[p][q] = 0;
      } 
  }
  reInitGameVar();
  
}

int currStartCol = 0;
int currStartPos[2] = {1,1};

const int startText[5][18] = {{1,1,1,4,0,0,5,5,5,6,0,6,0},
                              {1,0,1,4,0,0,5,0,5,6,0,6,0},
                              {1,1,1,4,0,0,5,5,5,0,6,0,0},
                              {1,0,0,4,0,0,5,0,5,0,6,0,0},
                              {1,0,0,4,4,4,5,0,5,0,6,0,0}};

const int pacmanClose[6][5] = { {0,5,5,5,0},
                                {5,5,5,5,5},
                                {5,5,5,5,5},
                                {5,5,5,5,5},
                                {5,5,5,5,5},
                                {0,5,5,5,0}};

const int pacmanOpen[6][5] = { {0,5,5,5,5},
                                {5,5,5,5,0},
                                {5,5,5,0,0},
                                {5,5,5,0,0},
                                {5,5,5,5,0},
                                {0,5,5,5,5}};                      

int PacmanPos[2] = {7,4}; //y,x

void loop() {
  if (!inGame){
    if (getBtn()){
        // escape the menu
        inGame = true;
        for (int p = 0; p < 13; p++)
          {
            copy(startPacMap[p], currScreen[p], 13);
          }
        setScreen(currScreen);
        playSong(pacman, musicPin, 62, 500);
        
      }
    else {
        //run animations
        //text
        for (int j = 0; j < 5; j++)
          {
            for (int i = 0; i < gridSize-2; i++) 
            {
                currScreen[currStartPos[0] + j][currStartPos[1] + i] = startText[j][(i + currStartCol)%18];
            }
          }
          currStartCol ++;

        //pacman
        if((int)floor(millis()/500)%2==0){
          //display closed
          for (int j = 0; j < 6; j++)
          {
            for (int i = 0; i < 5; i++) 
            {
                currScreen[PacmanPos[0] + j][PacmanPos[1] + i] = pacmanClose[j][i];
            }
          }
        }
        else{
          //display open
          for (int j = 0; j < 6; j++)
          {
            for (int i = 0; i < 5; i++) 
            {
                currScreen[PacmanPos[0] + j][PacmanPos[1] + i] = pacmanOpen[j][i];
            }
          }
        }
        
      }

    
   
    setScreen(currScreen);

    delay(175); // menu animation fps
  }
  
  /*
  █▀▀ ▄▀█ █▀▄▀█ █▀▀
  █▄█ █▀█ █░▀░█ ██▄*/
  
  else
  {
    //switch to pacmap
    for (int i = 0; i < gridSize; i++)
    {
      for (int j = 0; j < gridSize; j++)
      {
        currScreen[i][j] = pacMap[i][j]; 
      }  
    }

    delay(333); // change framerate


    //process ghosts
    for (int g = 0; g < numOfGhosts; g++) {
      //if not dead
      if (ghosts[g][5] == -1){
        ghostAi(ghosts[g][0], ghosts[g][1], ghosts[g][2], ghosts[g][3]);
        currScreen[ghosts[g][0]][ghosts[g][1]] = ghosts[g][4];
        //check if colliding with pacman
        if (ghosts[g][0] == pacPos[0] && ghosts[g][1] == pacPos[1]) {
          if (powerUpTime > 0) {
            ghosts[g][5] = 10; //frames until respawn after being eaten
          }
          else {
            die();
            break;
          }
        }
      }
      else if (ghosts[g][5] == 0) {
        //respawn
        int respawnY;
        int respawnX;
        //ypos
        if (pacPos[0] > 7){
          respawnY = 1;
        }
        else {
          respawnY = 11;
        }
        //xpos
        if (pacPos[1] > 7){
          respawnX = 1;
        }
        else {
          respawnX = 11;
        }

        ghosts[g][0] = respawnY;
        ghosts[g][1] = respawnX;

        ghosts[g][5] = -1;
      }
      else {
        ghosts[g][5] -= 1;
      }
    }
      //move pacman
      if (getAxis(yA) == 1 && checkCollide(pacPos[0] - 1, pacPos[1])) {
        //move up
        pacPos[0] -= 1;
      }
      else if (getAxis(yA) == -1 && checkCollide(pacPos[0] + 1, pacPos[1])) {
        //move down
        pacPos[0] += 1;
      }
      else if (getAxis(xA) == -1 && checkCollide(pacPos[0], pacPos[1] + 1)) {
        //move right
        pacPos[1] += 1;
      }
      else if (getAxis(xA) == 1 && checkCollide(pacPos[0], pacPos[1] - 1)) {
        //move left
        pacPos[1] -= 1;
      }
  
      //eat normal pellet
      if (pacMap[pacPos[0]][pacPos[1]] == 2) {
        pacMap[pacPos[0]][pacPos[1]] = 0;
      }
  
      //power pellet ♦
      if (pacMap[pacPos[0]][pacPos[1]] == 3) {
        pacMap[pacPos[0]][pacPos[1]] = 0;
        powerUpTime = 10;
      }
  
      if (powerUpTime > 0) {

      colors[9] = CRGB(255,0,255);
      colors[7] = CRGB(255,0,255);
      colors[8] = CRGB(255,0,255);
        tone(musicPin, 440, 10);
        powerUpTime -= 1;
      }
      else{
        colors[9] = CRGB(255,150,0);
        colors[7] = CRGB(255,0,150);
        colors[8] = CRGB(0,255,255);
      }
      //assign pacman
      if(inGame){
        currScreen[pacPos[0]][pacPos[1]] = 5;
      }
      if (levelDone()) {
        win();
      }
      
    
    
    setScreen(currScreen);
  }
}
